import re
from collections import OrderedDict
from collections.abc import Hashable
from urllib.parse import quote

from werkzeug.routing import parse_rule

from oapispec2.model import Model
from oapispec2.core.utils import merge, not_none



#: Maps Flask/Werkzeug rooting types to Swagger ones
PATH_TYPES = {
    'int': 'integer',
    'float': 'number',
    'string': 'string',
    'str': 'string',
    'default': 'string'
}


#: Maps Python primitives types to Swagger ones
PY_TYPES = {
    int: 'integer',
    float: 'number',
    str: 'string',
    bool: 'boolean',
    None: 'void'
}


class OpenApi:
    '''
    A Swagger documentation wrapper for an API instance.
    '''
    def __init__(self, metadata, handlers):
        self.metadata = metadata
        self.handlers = handlers

    def as_dict(self):
        return create_openapi_spec_dict(self.metadata, self.handlers)


def create_openapi_spec_dict(metadata, handlers):
    '''
    Output the specification as a serializable ``dict``.

    :returns: the full Swagger specification in a serializable format
    :rtype: dict
    '''

    paths = build_paths(handlers)
    models = find_models(handlers)

    return not_none({
        'swagger': '2.0',
        'basePath': parse_base_path(metadata.base_path),
        'paths': not_none(paths),
        'info': build_infos(metadata),
        'produces': metadata.representations,
        'consumes': ['application/json'],
        'securityDefinitions': metadata.authorizations or None,
        'security': security_requirements(metadata.security) or None,
        'tags': extract_tags(metadata, handlers),
        'definitions': serialize_definitions(models) or None,
        'host': metadata.host,
    })

def parse_base_path(base_path):
    if len(base_path) > 1 and base_path.endswith('/'):
        base_path = base_path[:-1]
    return base_path

def build_infos(metadata):
    infos = {
        'title': metadata.title,
        'version': metadata.version
    }

    if metadata.description:
        infos['description'] = metadata.description

    if metadata.terms_url:
        infos['termsOfService'] = metadata.terms_url

    if metadata.contact and (metadata.contact_email or metadata.contact_url):
        infos['contact'] = {
            'name': metadata.contact,
            'email': metadata.contact_email,
            'url': metadata.contact_url,
        }

    if metadata.license:
        infos['license'] = {'name': metadata.license}
        if metadata.license_url:
            infos['license']['url'] = metadata.license_url

    return infos

def build_paths(handlers):
    docs = [{ **getattr(handler, '__apidoc__', {}), 'handler': handler } for handler in handlers]
    paths = {}
    for apidoc in docs:
        route = apidoc.get('route')
        stripped_route = clean_route(route)

        current_route_obj = paths.get(stripped_route, {})
        current_route_obj[apidoc['method']] = apidoc
        paths[stripped_route] = current_route_obj

    for path, methods in paths.items():
        for method, doc in methods.items():
            paths[path][method] = serialize_operation(doc)

    return paths

def ref(model):
    '''Return a reference to model in definitions'''
    name = model.name if isinstance(model, Model) else model
    return {'$ref': '#/definitions/{0}'.format(quote(name, safe=''))}

def strip_route_types(route):
    '''Takes a route ('/path/to/<str:id>') and removes the type
    parameters ('/path/to/<id>')'''
    match = re.search(r'\/\<([a-z]+):', route)
    if not match:
        return route
    typ = match.groups()[0]
    route = route.replace(f'/<{typ}:', '/<')
    return strip_route_types(route)

def convert_path_args_to_brackets(route):
    '''Takes a route ('/path/to/<id>') and converts the
    arrows to brackets ('/path/to/{id}')'''
    match = re.search(r'\/\<([a-z_]+)\>', route)
    if not match:
        return route
    param = match.groups()[0]
    route = route.replace(f'/<{param}>', f'/{{{param}}}')
    return convert_path_args_to_brackets(route)

def clean_route(route):
    return convert_path_args_to_brackets(strip_route_types(route))

def extract_path_params(path):
    '''
    Extract Flask-style parameters from an URL pattern as Swagger ones.
    '''
    params = OrderedDict()
    for converter, _, variable in parse_rule(path):
        if not converter:
            continue
        param = {
            'name': variable,
            'in': 'path',
            'required': True
        }

        if converter in PATH_TYPES:
            param['type'] = PATH_TYPES[converter]
        else:
            raise ValueError(f'Unsupported type converter: {converter}')
        params[variable] = param
    return params

def create_header_object(header):
    if isinstance(header, str):
        header = {'description': header}
    typedef = header.get('type', 'string')
    if isinstance(typedef, Hashable) and typedef in PY_TYPES:
        header['type'] = PY_TYPES[typedef]
    elif isinstance(typedef, (list, tuple)) and len(typedef) == 1 and typedef[0] in PY_TYPES:
        header['type'] = 'array'
        header['items'] = {'type': PY_TYPES[typedef[0]]}
    elif hasattr(typedef, '__schema__'):
        header.update(typedef.__schema__)
    else:
        header['type'] = typedef
    return not_none(header)

def extract_tags(metadata, handlers):

    tags = []
    by_name = {}

    for tag in metadata.tags:
        if isinstance(tag, str):
            tag = {'name': tag}
        if isinstance(tag, dict) and 'name' in tag:
            tags.append(tag)
            by_name[tag['name']] = tag

    namespaces = [getattr(h, '__apidoc__', {}).get('namespace', {}) for h in handlers]

    for ns in namespaces:
        name = ns.get('name')
        if name not in by_name:
            tags.append(ns)
            by_name[name] = True

    return tags

def expected_params(apidoc):
    params = OrderedDict()

    for model, description in apidoc.get('expect', []):

        if isinstance(model, Model):
            params['payload'] = not_none({
                'name': 'payload',
                'required': True,
                'in': 'body',
                'schema': ref(model)
            })
            continue

        params['payload'] = not_none({
            'name': 'payload',
            'required': True,
            'in': 'body',
            'schema': ref(model),
            'description': description
        })

    return params

def get_operation_consumes(parameters):
    if not parameters:
        return None
    if not any(p['in'] == 'formData' for p in parameters):
        return None
    if any(p['type'] == 'file' for p in parameters):
        return ['multipart/form-data']
    return ['application/x-www-form-urlencoded', 'multipart/form-data']

def get_operation_namespace(apidoc):
    return [apidoc['namespace']['name']] if 'namespace' in apidoc else None

def get_operation_produces(apidoc):
    return [apidoc['produces']] if 'produces' in apidoc else None

def serialize_operation(apidoc):

    parameters = parameters_for(apidoc)

    operation = {
        'responses': responses_for(apidoc) or None,
        # 'summary': 'TODO-ray parse docstirng from here', # doc[method]['docstring']['summary'],
        'description': apidoc.get('description') or None,
        'operationId': apidoc.get('name'),
        'parameters': parameters,
        'security': security_for(apidoc),
        'consumes': get_operation_consumes(parameters),
        'produces': get_operation_produces(apidoc),
        'tags': get_operation_namespace(apidoc),
        'deprecated': True if apidoc.get('deprecated', False) else None
    }

    operation.update(vendor_fields(apidoc))

    return not_none(operation)

def vendor_fields(apidoc):
    '''
    Extract custom 3rd party Vendor fields prefixed with ``x-``

    See: http://swagger.io/specification/#specification-extensions-128
    '''
    return dict(
        (k if k.startswith('x-') else f'x-{k}', v)
        for k, v in apidoc.get('vendor', {}).items()
    )

def create_parameter(name, param):
    param['name'] = name

    if 'type' in param and 'schema' not in param:
        ptype = param.get('type', None)
        if isinstance(ptype, (type, type(None))) and ptype in PY_TYPES:
            param['type'] = PY_TYPES[ptype]

    return param

def parameters_for(apidoc):

    params = []

    route = apidoc.get('route')

    if route is None:
        raise ValueError('route value cannot be empty. Every function must be decorated with @doc.route("/route/path")')

    expected_paramaters = expected_params(apidoc)
    doc_params = apidoc.get('params', {})
    path_params = extract_path_params(route)

    all_params = merge(merge(expected_paramaters, doc_params), path_params)

    for name, param in all_params.items():
        param = create_parameter(name, param)
        params.append(param)

    return params if len(params) > 0 else None

def responses_for(apidoc):

    responses = {}
    global_headers = apidoc.get('headers', {})

    for code, (description, model, headers) in apidoc.get('responses', {}).items():

        description = description or 'Success'
        headers = headers or {}

        responses[code] = {'description': description}

        if model is not None:
            schema = ref(model)
            responses[code]['schema'] = schema

        responses[code] = attatch_headers(responses[code], global_headers, headers)

    return responses

def attatch_headers(response, global_headers, response_headers):
    all_headers = { **global_headers, **response_headers }
    if not all_headers:
        return response
    return {
        **response,
        'headers': dict((k, create_header_object(v)) for k, v in all_headers.items())
    }

def serialize_definitions(registered_models):
    return dict(
        (name, model.__schema__)
        for name, model in registered_models.items()
    )

def find_models(handlers):
    models = {}
    apidoc_list = [getattr(h, '__apidoc__') for h in handlers]
    for apidoc in apidoc_list:
        for expect, _ in apidoc.get('expect', []):
            if isinstance(expect, Model):
                models[expect.name] = expect
        for _, (_, model, _) in apidoc.get('responses', {}).items():
            if isinstance(model, Model):
                models[model.name] = model
    return models

def security_for(apidoc):
    security = apidoc.get('security', None)
    return security if security is None else security_requirements(security)

def security_requirements(value):
    if not value:
        return []
    return [{value: []}]



from oapispec2.core.utils import immutable, not_none
from oapispec2.model import Model


def _eval(value):
    '''Evaluates a value - returning it or its result if its callable'''
    return value() if callable(value) else value

def create_schema(
    type='object',
    default=None,
    format=None,
    title=None,
    description=None,
    readonly=None,
    example=None,
    **kwargs):

    required = kwargs.pop('required', False)
    discriminator = kwargs.pop('discriminator', None)

    schema = not_none({
        'type': type,
        'format': format,
        'title': title,
        'description': description,
        'readOnly': readonly,
        'default': _eval(default),
        'example': example,
        **kwargs
    })

    return immutable({
        '__schema__': schema,
        'schema': lambda: schema,
        'required': required,
        'discriminator': discriminator,
        'description': description
    })

def raw(**kwargs):
    return create_schema(**kwargs)

def nested(model, as_list=False, **kwargs):

    nested_name = getattr(model, 'resolved', model).name
    ref = { '$ref': f'#/definitions/{nested_name}' }

    return create_schema(
        type='array' if as_list else None,
        **ref,
        **kwargs)

def array(item_type, min_items=None, max_items=None, unique=None, **kwargs):
    if isinstance(item_type, Model):
        return create_schema(**item_type._schema)
    return create_schema(
        type='array',
        minItems=_eval(min_items),
        maxItems=_eval(max_items),
        uniqueItems=_eval(unique),
        items=item_type.schema(),
        **kwargs)

def string(enum=None, min_length=None, max_length=None, pattern=None, **kwargs):
    enum = _eval(enum)
    if enum and 'example' not in kwargs:
        kwargs['example'] = enum[0]

    # If a discriminator was provided
    # attempt to set required
    if kwargs.get('discriminator', False):
        # Set required True unless caller passed
        # it explicitly as False
        kwargs['required'] = True

    return create_schema(
        type='string',
        minLength=_eval(min_length),
        maxLength=_eval(max_length),
        pattern=_eval(pattern),
        enum=enum if enum else None,
        **kwargs)

def integer(minimum=None, exclusive_minimum=None, maximum=None, exclusive_maximum=None, **kwargs):
    return create_schema(
        type='integer',
        minimum=_eval(minimum),
        exclusiveMinimum=_eval(exclusive_minimum),
        maximum=_eval(maximum),
        exclusiveMaximum=_eval(exclusive_maximum),
        multipleOf=_eval(kwargs.get('multiple', None)),
        **kwargs)

def float(minimum=None, exclusive_minimum=None, maximum=None, exclusive_maximum=None, **kwargs):
    return create_schema(
        type='number',
        minimum=_eval(minimum),
        exclusiveMinimum=_eval(exclusive_minimum),
        maximum=_eval(maximum),
        exclusiveMaximum=_eval(exclusive_maximum),
        multipleOf=_eval(kwargs.get('multiple', None)),
        **kwargs)

def boolean(**kwargs):
    return create_schema(type='boolean', **kwargs)

def date_time(minimum=None, exclusive_minimum=None, maximum=None, exclusive_maximum=None, **kwargs):
    return create_schema(
        type='string',
        format='date-time',
        minimum=_eval(minimum),
        exclusiveMinimum=_eval(exclusive_minimum),
        maximum=_eval(maximum),
        exclusiveMaximum=_eval(exclusive_maximum),
        **kwargs)

def date(minimum=None, exclusive_minimum=None, maximum=None, exclusive_maximum=None, **kwargs):
    return create_schema(
        type='string',
        format='date',
        minimum=_eval(minimum),
        exclusiveMinimum=_eval(exclusive_minimum),
        maximum=_eval(maximum),
        exclusiveMaximum=_eval(exclusive_maximum),
        **kwargs)



import inspect
import datetime

from oapispec2.model import Model
from oapispec2 import fields


def get_annotations(klass):
    return getattr(klass, '__annotations__', {})

def get_model_attribute_type(attr_type):
    if attr_type == str:
        return fields.string()
    if attr_type == bool:
        return fields.boolean()
    if attr_type == dict:
        return fields.raw()
    if attr_type == int:
        return fields.integer()
    if attr_type == datetime.date:
        return fields.date()
    if attr_type == datetime.datetime:
        return fields.date_time()

    type_str = str(type(attr_type))
    is_list = 'Generic' in type_str

    if is_list:
        child_type = attr_type.__args__[0]
        return fields.array(get_model_attribute_type(child_type))

    return fields.nested(build_model(attr_type))

def get_model_attribute_dict(attribute_map):
    return dict([
        (name, get_model_attribute_type(attr_type)) for name, attr_type in attribute_map.items()
    ])

def build_model(klass):
    members = get_annotations(klass)
    return Model(klass.__name__, get_model_attribute_dict(members))



import os


def generate_swagger_ui(
    spec_url,
    title='Swagger UI',
    asset_source='https://cdn.jsdelivr.net/npm/swagger-ui-dist@3.25.0'):

    ui_template = load_ui_template()

    config = {
        'title': title,
        'asset_source': asset_source,
        'spec_url': spec_url
    }

    for key, value in config.items():
        ui_template = ui_template.replace('{{%s}}' % key, value)

    return ui_template

def load_ui_template():
    rel_path = '../templates/swagger-ui.html'
    path = os.path.join(os.path.dirname(__file__), rel_path)
    with open(path) as f:
        return f.read()




from oapispec2.core.openapi import OpenApi
from oapispec2.core.utils import immutable
from oapispec2.core.swagger import generate_swagger_ui


def schema(handlers=None, metadata=None):

    handlers = handlers or []
    metadata = metadata or {}
    metadata = meta(**metadata)

    def register(handler):
        return schema(handlers=[*handlers, handler], metadata=metadata)

    def generate():
        return OpenApi(metadata, handlers).as_dict()

    def generate_ui(spec_url):
        '''Generates the swagger-ui html file and returns the content
        :param str spec_url: The url swagger-ui should use to load your valid OpenAPI spec. Ex. spec_url='http://myapp.io/swagger-spec.json'
        '''
        return generate_swagger_ui(spec_url, title=metadata.title)

    return immutable(dict(
        register=register,
        generate=generate,
        handlers=handlers,
        metadata=metadata,
        generate_ui=generate_ui
    ))


def meta(
    version='1.0',
    title='API',
    host=None,
    description=None,
    terms_url=None,
    contact=None,
    contact_url=None,
    contact_email=None,
    license=None,
    license_url=None,
    authorizations=None,
    security=None,
    tags=None,
    base_path='/',
    representations=None):

    return immutable(dict(
        version=version,
        title=title,
        host=host,
        description=description,
        terms_url=terms_url,
        contact=contact,
        contact_url=contact_url,
        contact_email=contact_email,
        license=license,
        license_url=license_url,
        authorizations=authorizations,
        security=security,
        tags=tags or [],
        base_path=base_path,
        representations=representations or ['application/json']
    ))



import re

from http import HTTPStatus
from collections import OrderedDict
from copy import deepcopy

from oapispec2.core.immutable import Immutable


def immutable(obj=None, **kwargs):
    if obj is not None:
        return Immutable(**obj)
    return Immutable(**kwargs)

def merge(first, second):
    '''
    Recursively merges two dictionaries.

    Second dictionary values will take precedence over those from the first one.
    Nested dictionaries are merged too.

    :param dict first: The first dictionary
    :param dict second: The second dictionary
    :return: the resulting merged dictionary
    :rtype: dict
    '''
    result = deepcopy(first)
    for key, value in second.items():
        if key in result and isinstance(result[key], dict):
            result[key] = merge(result[key], value)
        else:
            result[key] = deepcopy(value)
    return result


def not_none(data):
    '''
    Remove all keys where value is None

    :param dict data: A dictionary with potentially some values set to None
    :return: The same dictionary without the keys with values to ``None``
    :rtype: dict
    '''
    return dict(OrderedDict((k, v) for k, v in sorted(data.items()) if v is not None))



import json

class Immutable:

    def __init__(self, **kwargs):
        """Sets all values once given
        whatever is passed in kwargs
        """
        for k,v in kwargs.items():
            object.__setattr__(self, k, v)

    def __setattr__(self, *args):
        """Disables setting attributes via
        item.prop = val or item['prop'] = val
        """
        raise TypeError('Immutable objects cannot have properties set')

    def __delattr__(self, *args):
        """Disables deleting properties"""
        raise TypeError('Immutable objects cannot have properties deleted')

    def __getitem__(self, item):
        """Allows for dict like access of properties
        val = item['prop']
        """
        return self.__dict__[item]

    def keys(self):
        """Paired with __getitem__ supports **unpacking
        new = { **item, **other }
        """
        return self.__dict__.keys()

    def get(self, *args, **kwargs):
        """Allows for dict like property access
        item.get('prop')
        """
        return self.__dict__.get(*args, **kwargs)

    def pprint(self):
        """Helper method used for printing that
        formats in a dict like way
        """
        def deb(o):
            return getattr(o, '__dict__', 'DICT NOT FOUND')
        return json.dumps(self,
            default=deb,
            sort_keys=True,
            indent=4)

    def __repr__(self):
        """Print to repl in a dict like fashion"""
        return self.pprint()

    def __str__(self):
        """Convert to a str in a dict like fashion"""
        return self.pprint()

    def dict(self):
        """Helper method for getting the raw dict value
        of the immutable object"""
        return self.__dict__

    def __eq__(self, other):
        """Supports equality operator
        immutable({'a': 2}) == immutable({'a': 2})"""
        if other is None:
            return False
        return self.dict() == other.dict()



import setuptools

from oapispec2.version import VERSION

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='oapispec2',
    version=VERSION,
    author='Ray Epps',
    author_email='rayharryepps@gmail.com',
    description='A library of decorators and functions for generating open api specs',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/dhakiki/oapispec',
    packages=[
        'oapispec2',
        'oapispec2.core'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Operating System :: OS Independent'
    ],
    python_requires='>=3',
    install_requires=[
        'aniso8601',
        'attrs',
        'importlib-metadata',
        'jsonschema',
        'more-itertools',
        'pyrsistent',
        'pytz',
        'six',
        'Werkzeug'
    ]
)



import re

# from jsonschema import Draft4Validator
# from jsonschema.exceptions import ValidationError

from oapispec2.core.utils import not_none


RE_REQUIRED = re.compile(r'u?\'(?P<name>.*)\' is a required property', re.I | re.U)

def _format_error(error):
    path = list(error.path)
    if error.validator == 'required':
        name = RE_REQUIRED.match(error.message).group('name')
        path.append(name)
    key = '.'.join(str(p) for p in path)
    return key, error.message

class Model:
    '''
    Handles validation and swagger style inheritance for both subclasses.
    Subclass must define `schema` attribute.

    :param str name: The model public name
    '''

    def __init__(self, name, attributes):
        self.attributes = attributes

        self.__apidoc__ = {
            'name': name
        }
        self.name = name
        self.__parents__ = []

    @property
    def __schema__(self):
        schema = self._schema

        if self.__parents__:
            refs = [
                {'$ref': '#/definitions/{0}'.format(parent.name)}
                for parent in self.__parents__
            ]

            return {
                'allOf': refs + [schema]
            }
        return schema

    def inherit(self, name, attributes):
        '''
        Inherit this model (use the Swagger composition pattern aka. allOf)
        :param str name: The new model name
        :param dict fields: The new model extra fields
        '''
        model = Model(name, attributes)
        model.__parents__ = [*self.__parents__, self]
        return model

    # def validate(self, data):
    #     validator = Draft4Validator(self.__schema__)
    #     try:
    #         validator.validate(data)
    #     except ValidationError:
    #         return dict(_format_error(e) for e in validator.iter_errors(data))
    #     return None

    def __str__(self):
        return 'Model({name},{{{fields}}})'.format(name=self.name, fields=','.join(self.attributes.keys()))

    @property
    def _schema(self):
        properties = {}
        required = set()
        discriminator = None
        for name, field in self.attributes.items():
            properties[name] = field.__schema__
            if field.required:
                required.add(name)
            if getattr(field, 'discriminator', False):
                discriminator = name

        return not_none({
            'required': sorted(list(required)) or None,
            'properties': properties,
            'discriminator': discriminator,
            'type': 'object',
        })



'''The doc module provides decorator functions that will modify a
functions __apidoc__ object. For example, when using `method('GET')`
it will set the value of the functions method as 'get' on its __apidoc__'''

from http import HTTPStatus

from oapispec2.core.utils import merge, not_none


def doc(**kwargs):
    '''A decorator to add documentation metadata to the decorated function'''
    def wrapper(documented):
        current_doc = getattr(documented, '__apidoc__', {})
        if 'name' not in current_doc:
            kwargs['name'] = documented.__name__
        documented.__apidoc__ = merge(current_doc, kwargs)
        return documented
    return wrapper

def route(endpoint):
    '''
    A decorator to route resources.
    '''
    return doc(route=endpoint)

def hide(func):
    '''A decorator to hide a resource or a method from specifications'''
    return doc()(func)

def expect(model, description=None):
    '''
    A decorator to Specify the expected input model

    :param ModelBase inputs: An expect model
    :param bool validate: whether to perform validation or not

    '''
    return doc(expect=[(model, description)])

def param(name, type=str, location='query', description=None, format=None, required=False, default=None):
    '''
    A decorator to specify one of the expected parameters

    :param str name: the parameter name
    :param str description: a small description
    :param str location: the parameter location `(query|header|formData|body|cookie)`
    '''
    return doc(params={
        name: not_none({
            'in': location,
            'type': type,
            'format': format,
            'default': default,
            'required': True if required else None,
            'description': description
        })
    })

def response(http_status: HTTPStatus, model=None, headers=None):
    '''
    A decorator to specify one of the expected responses

    :param HTTPStatus http_status: the HTTP status (or any object with a 'value' and 'description' property)
    :param ModelBase model: an optional response model
    :param dict<str, header> headers: an optional dict of headers that are returned with this response
    '''
    code = http_status.value
    description = http_status.description
    return doc(responses={str(code): (description, model, headers)})

def header(name, description=None, **kwargs):
    '''
    A decorator to specify one of the expected headers

    :param str name: the HTTP header name
    :param str description: a description about the header

    '''
    return doc(headers={
        name: {
            'description': description
        },
        **kwargs
    })

def produces(mimetypes):
    '''A decorator to specify the MIME types the API can produce'''
    return doc(produces=mimetypes)

def deprecated(func):
    '''A decorator to mark a resource or a method as deprecated'''
    return doc(deprecated=True)(func)

def vendor(*args, **kwargs):
    '''
    A decorator to expose vendor extensions.

    Extensions can be submitted as dict or kwargs.
    The ``x-`` prefix is optionnal and will be added if missing.

    See: http://swagger.io/specification/#specification-extensions-128
    '''
    for arg in args:
        kwargs.update(arg)
    return doc(vendor=kwargs)

def method(http_method):
    '''A decorator to set the method for the handler'''
    return doc(method=http_method.lower())

def namespace(name, description=None):
    '''A decorator that groups the decorated handler function in a namespace'''
    return doc(namespace=not_none({
        'name': name,
        'description': description
    }))
